﻿Option Strict Off
Option Explicit On
Imports MSCommLib
Imports System.IO
Imports System.IO.Ports
Imports System.Drawing.Drawing2D
Imports System.Math

Public Class MainForm
    Const SampleRate As Double = 1.25               'Im PIC festgelegte Samplerate in ms
    Const MBTXT0 As String = "-16A bis +16A"
    Const MWSTG0 As Double = 1 / 40                 'Steigung im Messbereich 0: Experimentell ermittelt 40 ADC-Werte/Ampere
    Const MWOFF0 As Double = -12.8                  'Offset im Messbereich 0: Rechnerisch ermittelt, so dass ADC=512 bei 0 Ampere liegt - Offset muss eh im Betrieb regelmaessig genullt werden
    Const MBTXT1 As String = "-20A bis +20A"
    Const MWSTG1 As Double = 1 / 25.9               'Steigung im Messbereich 1: Experimentell ermittelt 25,9 ADC-Werte/Ampere
    Const MWOFF1 As Double = -19.8                  'Offset im Messbereich 1: Rechnerisch ermittelt, so dass ADC=512 bei 0 Ampere liegt - Offset muss eh im Betrieb regelmaessig genullt werden
    Const MBTXT2 As String = "-6A bis +97A"
    Const MWSTG2 As Double = 1 / 9.9                'Steigung im Messbereich 2: Experimentell ermittelt 9.9 ADC-Werte/Ampere
    Const MWOFF2 As Double = -6.6                   'Offset im Messbereich 2: Rechnerisch ermittelt, so dass ADC=65 bei 0 Ampere liegt - Offset muss eh im Betrieb regelmaessig genullt werden
    Const MBTXT3 As String = "-8A bis +143A"
    Const MWSTG3 As Double = 1 / 6.8                'Steigung im Messbereich 3: Experimentell ermittelt 6,8 ADC-Werte/Ampere
    Const MWOFF3 As Double = -7.8                   'Offset im Messbereich 3: Rechnerisch ermittelt, so dass ADC=53 bei 0 Ampere liegt - Offset muss eh im Betrieb regelmaessig genullt werden

    Public WithEvents MSComm1 As New MSComm
    Dim MWStg, MWOff As Double                      'Berechnungskoeffizienten um Integer vom Gerät in Ampere umzurechnen (abhängig von eingestelltem Mess-Modus)
    Dim AktAvgI As Double                           'Aktueller Average Messwert in Ampere - nötig, um jederzeit den Anzeigebereich nullen zu können
    Dim AktRohWerte() As UInt16                     'Aktuell aktive Rohwerte (entweder vom PIC empfangen oder aus Fielstream - nötig, um jederzeit die Werte abspeichern oder neu anzeigen zu können
    Dim fstream() As UInt16                         'Datastream fuer Connection mit File
    Dim fspointer As Integer                        'Pointer in Filestream
    Dim messzeit As Double                          'Eingestellte Messzeit
    Dim connected As Boolean
    Dim imageshow As Boolean

    'Spezielle Variablen der Anzeige
    Const RANDL As Byte = 1             'Linker Rand in Anzeige
    Const RANDR As Byte = 1             'Rechter Rand in Anzeige
    Const RANDO As Byte = 6             'Oberer Rand in Anzeige
    Const RANDU As Byte = 6             'Unterer Rand in Anzeige
    Const SKYABSTMIN As Byte = 30       'Minimaler Abstand der Y-Skalenlinien in Pixel
    Const SKXABSTMIN As Byte = 50       'Minimaler Abstand der X-Skalenlinien in Pixel
    Dim YLbl() As Label                 'Labels der Y-Skala
    Dim XLbl() As Label                 'Labels der X-Skala
    Dim PixProYVal As Double            'y-Skalierung Pixel/Value
    Dim Ynull As Integer                'Die y-Koordinate der Anzeige, bei der die Null-Linie liegt (liegt evtl außerehalb der Anzeige, wenn z.B. nur ein positiver oder nur ein negativer Bereich angezeigt werden soll)


    Private Sub MainForm_Load(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Load
        FillCOMPortList()
        C_COMPorts.Enabled = True
        B_Connect.Text = "Connect"
        B_Connect.Enabled = True
        connected = False

        C_Modus.Enabled = False
        C_Modus.Items.Add(MBTXT0)
        C_Modus.Items.Add(MBTXT1)
        C_Modus.Items.Add(MBTXT2)
        C_Modus.Items.Add(MBTXT3)
        C_Modus.SelectedIndex = 0

        C_Timebase.Enabled = False
        C_Timebase.Items.Add("100")
        C_Timebase.Items.Add("350")
        C_Timebase.Items.Add("700")
        C_Timebase.Items.Add("2000")
        C_Timebase.Items.Add("5000")
        C_Timebase.Items.Add("6000")
        C_Timebase.Items.Add("8000")
        C_Timebase.SelectedIndex = 0
        messzeit = Convert.ToInt32(C_Timebase.Text) / 1000

        B_Load.Enabled = True
        B_Save.Enabled = True
        Timer.Enabled = False

        'Dummy-Initialisierung des Label-Arrays. Ist nötig, damit erstes dispose in MakeSkale nicht ins leere läuft
        ReDim YLbl(1)
        YLbl(0) = New Label
        ReDim XLbl(1)
        XLbl(0) = New Label

        AktAvgI = 0
        imageshow = False

        'Ab hier Testdatenerzeugung - auskommentieren, wenn gewünscht
        'Dim x As Integer
        'Dim F As Double
        'Dim AnzWerte As Integer
        'MWStg = 1 / 50              'Koeffizienten, so dass Ströme von -10 bis +10 Amp auf den Bereich 0-1000 abgebildet werden
        'MWOff = -10
        'AnzWerte = 500
        'ReDim AktRohWerte(AnzWerte - 1)                '-1 weil Redim nicht auf length sondern auf MaxIndex arbeitet
        'For x = 0 To AnzWerte - 1
        '    F = 10 * Math.Sin(4 * 2 * Math.PI * x / 500)            'Werte von 0 bis 10A
        '    AktRohWerte(x) = Math.Round((F - MWOff) / MWStg)          'Analoge Ströme umrechnen in 16Bit mit MW Koeffizienten (so dass 16-Bit-Werte von 0 bis 1000 liegen), PIC berechnet V aus ADC-Werten: V = (x*Stg/256 + Off)/2 - Ergebnis ist dann in 1/100V - das hier ist die Rückrechnung von 1V in Rohwert
        'Next x
        'Display_StandardMake(AktRohWerte, AktRohWerte.Length * SampleRate / 1000)
        'imageshow = True
        'Ende der Testdatenerzeugung

    End Sub

    Private Sub B_Connect_Click(ByVal sender As Object, ByVal e As EventArgs) Handles B_Connect.Click
        If B_Connect.Text = "Connect" Then
            Connect()
            B_Load.Enabled = False
            B_Save.Enabled = False
            B_Connect.Text = "Disconnect"
            C_COMPorts.Enabled = False
            C_Modus.Enabled = True
            C_Timebase.Enabled = True
        Else
            Disconnect()
            B_Load.Enabled = True
            B_Save.Enabled = True
            B_Connect.Text = "Connect"
            C_COMPorts.Enabled = True
            C_Modus.Enabled = False
            C_Timebase.Enabled = False
        End If
    End Sub

    Private Sub Connect()
        'Verbindet mit File oder COM-Port
        Dim RXlength As Integer                         'Durch die eingestellte Zeitbasis vorgegebene Messwert-Anzahl

        If C_COMPorts.SelectedIndex = 0 Then            'Connect mit File
            Dim Myfile As String = ""
            Dim reader As StreamReader
            Dim inputstr As String
            Dim n As Integer

            With OpenFileDialog
                .Filter = "Messzangen-Files (*.kmz)|*.kmz"
                .FilterIndex = 1
                '.InitialDirectory = OpStrings(1)
                .Title = "Messzangen-File laden"
                .FileName = ""
                If .ShowDialog() = DialogResult.OK Then Myfile = .FileName
            End With
            If Len(Myfile) > 0 Then
                Try
                    reader = New StreamReader(Myfile)
                Catch
                    MsgBox("Ladefehler", vbOK + vbCritical, "Fehler")
                    Exit Sub
                End Try

                Try
                    reader.BaseStream.Seek(0, SeekOrigin.Begin)     ' Pointer auf file-Anfang
                    reader.ReadLine()                               'Erkennungszeile
                    MWStg = CDbl(reader.ReadLine())               'Berechnungskoeffizienten
                    MWOff = CDbl(reader.ReadLine())
                    n = 0
                    While Not reader.EndOfStream
                        ReDim Preserve fstream(n)
                        inputstr = reader.ReadLine
                        fstream(n) = CInt(inputstr)
                        n = n + 1
                    End While
                    fspointer = 0
                    Timer.Enabled = True                        'Update der Anzeige Timer-basiert
                Catch
                    MsgBox("Lesefehler", vbOK + vbCritical, "Fehler")
                    reader.Close()
                    Exit Sub
                End Try
            End If
        Else                                            'Connect mit COM-Port
            RXlength = Math.Round(messzeit / SampleRate * 1000) * 2 + 1     '*2 weil zwei Byte pro Messwert gelesen werden, +1 um Sync auf High-Byte zu erlauben
            With MSComm1
                .CommPort = Val(Mid(C_COMPorts.SelectedItem, 4))
                .Settings = "19200,N,8,1"
                .RThreshold = RXlength                  'comEvReceive-Event wird ausgeloest wenn RXlength Werte angekommen sind
                .InputLen = RXlength                    'Input liest RXlength Werte
                .PortOpen = True
            End With
            connected = True
        End If
        imageshow = False
    End Sub

    Private Sub Disconnect()
        If C_COMPorts.SelectedIndex = 0 Then        'Connection mit File
            Timer.Enabled = False
        Else                                        'Connection mit COM-Port
            MSComm1.PortOpen = False
            connected = False
        End If
        imageshow = True
    End Sub

    Private Sub C_Modus_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles C_Modus.SelectedIndexChanged
        Dim ch As Byte

        ch = C_Modus.SelectedIndex
        If connected Then SetPICChannel(ch) 'Scahltet im PIC entsprechend die Widerstände um

        'Koeffizienten für die Berechnung Amp = MWStg*ADCwert + MWOff
        If ch = 0 Then              'Messbereich 1: 'ADC-Messbereich 0-1023 wird in den Ampere-Bereich von ca. -13 bis +13 umgerechnet
            MWStg = MWSTG0
            MWOff = MWOFF0
        ElseIf ch = 1 Then          'Noch leer - zeigt aktuell noch Rohwerte an
            MWStg = MWSTG1
            MWOff = MWOFF1
        ElseIf ch = 2 Then          'Noch leer - zeigt aktuell noch Rohwerte an
            MWStg = MWSTG2
            MWOff = MWOFF2
        Else                        'Messbereich Anlasserstrom: 'ADC-Messbereich 0-1023 wird in den Ampere-Bereich -8 bis +143 umgerechnet
            MWStg = MWSTG3
            MWOff = MWOFF3
        End If

    End Sub

    Private Sub C_Timebase_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles C_Timebase.SelectedIndexChanged
        Dim time As Integer
        Dim RXlength As Integer                         'Durch die eingestellte Zeitbasis vorgegebene Messwert-Anzahl

        time = Convert.ToInt32(C_Timebase.Text)         'Eingestellte Messzeit in ms
        messzeit = time / 1000                          'Messzeit in ms
        RXlength = Math.Round(time / SampleRate) * 2 + 1     '*2 weil zwei Byte pro Messwert gelesen werden, +1 um Sync auf High-Byte zu erlauben
        MSComm1.RThreshold = RXlength                   'Ab sofort neue RXlength empfangen
        MSComm1.InputLen = RXlength
    End Sub

    Private Sub B_Load_Click(ByVal sender As Object, ByVal e As EventArgs) Handles B_Load.Click
        Dim Myfile As String = ""
        Dim paket(8) As Char
        Dim n As Integer
        Dim reader As StreamReader
        Dim inputstr As String
        Dim enc As System.Text.Encoding = System.Text.Encoding.Default

        With OpenFileDialog
            .Filter = "Messzangen Files (*.kmz)|*.kmz"
            .FilterIndex = 1
            '.InitialDirectory = OpStrings(1)
            .Title = "Messzangen File laden"
            .FileName = ""
            If .ShowDialog() = DialogResult.OK Then Myfile = .FileName
        End With
        If Len(Myfile) > 0 Then
            Try
                reader = New StreamReader(Myfile)
            Catch
                MsgBox("Ladefehler", vbOK + vbCritical, "Fehler")
                Exit Sub
            End Try

            Try
                reader.BaseStream.Seek(0, SeekOrigin.Begin)     ' Pointer auf file-Anfang
                reader.ReadLine()                               'Erkennungszeile
                MWStg = CDbl(reader.ReadLine())               'Berechnungskoeffizienten
                MWOff = CDbl(reader.ReadLine())
                n = 0
                While Not reader.EndOfStream
                    ReDim Preserve AktRohWerte(n)
                    inputstr = reader.ReadLine
                    AktRohWerte(n) = CInt(inputstr)
                    n = n + 1
                End While
                Display_StandardMake(AktRohWerte, AktRohWerte.Length * SampleRate / 1000)
                imageshow = True
            Catch
                MsgBox("Lesefehler", vbOK + vbCritical, "Fehler")
                reader.Close()
                Exit Sub
            End Try
        End If
    End Sub

    Private Sub B_Save_Click(ByVal sender As Object, ByVal e As EventArgs) Handles B_Save.Click
        'Schreibt aktuelle Daten in ein .kmz-File
        Dim Myfile As String = ""
        Dim n As Integer

        With SaveFileDialog
            .Filter = "Messzangen-Files (*.kmz)|*.kmz"
            .FilterIndex = 1
            '.InitialDirectory = OpStrings(1)
            .Title = "Daten in Messzangen-File speichern"
            .FileName = ""
        End With
        If SaveFileDialog.ShowDialog() = DialogResult.OK Then
            Myfile = SaveFileDialog.FileName
            Dim quelle As New FileStream(Myfile, FileMode.Create)
            Dim writer As New StreamWriter(quelle)
            writer.BaseStream.Seek(0, SeekOrigin.Begin)     ' Pointer auf file-Anfang
            writer.WriteLine("1.0:KMZ - Messzangen-File")
            writer.WriteLine(MWStg.ToString)               'Berechnungskoeffizienten
            writer.WriteLine(MWOff.ToString)
            'Fuer HistSize Werte, werden nun die Messwerte nacheinander in das File geschrieben
            For n = 0 To AktRohWerte.Length - 1
                writer.WriteLine(AktRohWerte(n).ToString)
            Next
            writer.Close()
            quelle.Close()
        End If
    End Sub

    Private Sub FillCOMPortList()
        Dim ports As String() = SerialPort.GetPortNames()
        Dim port As String

        C_COMPorts.Items.Add("File")
        For Each port In ports                  'Alle COM-Ports einfuegen
            C_COMPorts.Items.Add(port)
        Next port
        If C_COMPorts.Items.Count > 1 Then
            C_COMPorts.SelectedIndex = 1
        Else
            C_COMPorts.SelectedIndex = 0
        End If
    End Sub

    Public Sub SetPICChannel(ByVal ch As Byte)
        'ch geht von 0 bis 3 - 2Bit, Low-Bit liegt auf Bit5 des Kommandos, High-Bit liegt auf Bit4 auf Bit4
        Dim outstr As String
        Dim cmd As Byte
        Dim flush As String
        Dim RXlength As Integer

        cmd = 0
        If ch = 1 Or ch = 3 Then cmd = cmd + 32 'Low-Bit (Bit 5 in cmd setzen)
        If ch = 2 Or ch = 3 Then cmd = cmd + 16 'High-Bit (bit 4 in cmd setzen)
        outstr = Chr(cmd)
        MSComm1.Output = outstr

        RXlength = MSComm1.InputLen             'Aktuelle Inputlen sichern
        Threading.Thread.Sleep(5)               'Eine Uebertragung an den PIC kann eine Uebertragung vom PIC unterbrechen und damit die Uebertragung kurzzeitig korrumpieren - 5ms warten um sicherzustellen, dass korrumpierte Uebertragung beendet ist...
        MSComm1.InputLen = 0                    '...dann gesamten Input-Buffer flushen
        flush = MSComm1.Input
        MSComm1.InputLen = RXlength             'Lese-Laenge wieder zurueck auf vorherige Einstellung
    End Sub

    Private Sub MSComm1_OnComm() Handles MSComm1.OnComm
        Dim Inputstr As String

        Select Case MSComm1.CommEvent
            Case MSCommLib.CommEventConstants.comEventOverrun
                Disconnect()
                MsgBox("comEventOverrun aufgetreten")
            Case MSCommLib.CommEventConstants.comEventRxOver
                Disconnect()
                MsgBox("comEventRxOver aufgetreten")
            Case MSCommLib.OnCommConstants.comEvReceive
                Dim enc As System.Text.Encoding = System.Text.Encoding.Default
                Dim n, add As Integer
                Dim highbyte, lowbyte As Byte
                Inputstr = MSComm1.Input                'Es werden die in C_Timebase_Change festgelegte Anzahl Bytes empfangen
                Dim bytearray(Len(Inputstr)) As Byte
                bytearray = enc.GetBytes(Inputstr)      'Inputstr in Byte-Array umwandeln
                ReDim AktRohWerte(((bytearray.Length - 1) / 2) - 1) '-1 weil ein zusätzliches Byte zur Highbyte-Synchronisierung empfangen wurde, /2 weil zwei empfangene Byte zu einem 16-Bit Messwert umgerechnet werden, -1 weil Redim nicht auf Length sondern auf MaxIndex arbeitet
                If bytearray(0) >= 128 Then             'Highbyte-Synchronisierung: Ist erstes empfangenes Byte ein High-Byte (MSB=1)?
                    add = 0                             'Ja: Dann bei Auswertung mit erstem Byte beginnen und letztes Byte verwerfen (es wurden Len*2+1 Byte empfangen)
                Else
                    add = 1                             'Nein: Dann erstes Byte ignorieren, Auswertung mit zweitem Byte beginnen (es wurden Len*2+1 Byte empfangen)
                End If
                For n = 0 To AktRohWerte.Length - 1
                    highbyte = bytearray(2 * n + add)
                    lowbyte = bytearray(2 * n + add + 1)
                    If highbyte < 128 Or lowbyte >= 128 Then        'Jedes Byte-Paar auf Kennzeichen-Bits prüfen (Highbyte mus MSB=1, Lowbyte muss MSB=0 haben)
                        highbyte = 255                              'Fehlerfall kennzeihnhen durch Ergebnis 0xFFFF
                        lowbyte = 255
                    Else
                        highbyte = highbyte - 128                   'Highbyte hat immer MSB=1 als Kennzeichnung des Highbytes -> Kennzeichnungs-Bit löschen
                        'LSB des Highbyte soll MSB des Lowbytes werden, Highbyte muss um ein Bit nach rechts geschiftet werden
                        If highbyte Mod 2 = 0 Then                  'Ist LSB des Highbyte=0?
                            highbyte = highbyte / 2                 'Ja: MSB des Lowbyte bleibt 0, Highbyte um ein Bit nach rechts shiften 
                        Else
                            lowbyte = lowbyte + 128                 'Nein: LSB des Highbyte ist 1 - MSB des Lowbyte setzen
                            highbyte = (highbyte - 1) / 2           'LSB des Highbyte löschen und Highbyte um ein Bit nach rechts shiften 
                        End If
                    End If
                    AktRohWerte(n) = highbyte * 256 + lowbyte
                Next
                Display_StandardMake(AktRohWerte, messzeit)
        End Select
    End Sub

    Private Sub Timer_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Timer.Tick
        'Timer zur Anzeige einer Connection mit einem File
        Dim n As Integer
        Dim Anzlen As Integer

        'Erzeugt Messwerte-Array durch Auffüllen aus dem Filestream
        Anzlen = Math.Round(messzeit / SampleRate * 1000)     'Je nach eingestellter Messzeit, Anzahl der anzuzeigenden Messwert berechnen
        ReDim AktRohWerte(Anzlen - 1)                       '-1 weil Redim nicht auf length sondern auf MaxIndex arbeitet
        For n = 0 To AktRohWerte.Length - 1
            AktRohWerte(n) = fstream(fspointer)
            fspointer = fspointer + 1
            If fspointer = fstream.Length Then fspointer = 0
        Next
        Display_StandardMake(AktRohWerte, messzeit)
    End Sub

    Private Function ROHtoAMP(ByRef ADC() As UInt16) As Double()
        Dim AMP() As Double
        Dim AMPSum As Double
        Dim GlatSum As Integer
        Dim ADCDiff, ADCFiktWert As Integer
        Dim n, x As Integer

        ReDim AMP(ADC.Length - 1)       '-1 weil Redim nicht auf length sondern auf MaxIndex arbeitet

        'Y-Value Double-Array generieren, Durchschnitt berechnen
        AMPSum = 0
        For n = 0 To ADC.Length - 1
            If S_Glaettung.Value = 0 Then           'Ohne Glättung: d.h. Werte nur mit Steigung und Offset umrechnen
                AMP(n) = ADC(n) * MWStg + MWOff
            Else                                    'Mit Glättung: Durchschnittswerte über den eingestellten Glättbereich ermitteln
                GlatSum = 0
                If n < S_Glaettung.Value Then       'Für die ersten Werte müssen zusätzliche "Vorab-Werte" extrapoliert werden
                    ADCDiff = CInt(ADC(1)) - CInt(ADC(0))   'CInts sind nötig weil ADC-Werte UInt sind, d.h. bei negativem Ergebnis würde es zu einer Exception kommen
                    ADCFiktWert = ADC(0)
                    For x = n - S_Glaettung.Value To -1      'Die Werte mit "Index<0" werden extrapoliert (linear die Steigung von Wert(0) zu Wert(1) zurückgeführt)
                        ADCFiktWert = ADCFiktWert - ADCDiff
                        GlatSum = GlatSum + ADCFiktWert
                    Next
                    For x = 0 To n + S_Glaettung.Value      'Die Werte ab Index=0 werden genommen wie sie sind
                        GlatSum = GlatSum + ADC(x)
                    Next
                ElseIf n < ADC.Length - S_Glaettung.Value Then  'Im mittleren Bereich werden der aktuelle Wert, Glättanzahl Werte vorher und Glättanzahl Werte danach zu einem Durchschnitt verrechnet
                    For x = n - S_Glaettung.Value To n + S_Glaettung.Value
                        GlatSum = GlatSum + ADC(x)
                    Next
                Else                                                'Für die letzten Werte müssen zusätzliche "Folge-Werte" extrapoliert werden
                    For x = n - S_Glaettung.Value To ADC.Length - 1      'Die Werte bis Index=length-1 werden genommen wie sie sind
                        GlatSum = GlatSum + ADC(x)
                    Next
                    ADCDiff = CInt(ADC(ADC.Length - 1)) - CInt(ADC(ADC.Length - 2)) ''CInts sind nötig weil ADC-Werte UInt sind, d.h. bei negativem Ergebnis würde es zu einer Exception kommen
                    ADCFiktWert = ADC(ADC.Length - 1)
                    For x = ADC.Length To n + S_Glaettung.Value      'Die Werte mit "Index>=Length" werden extrapoliert (linear die Steigung von Wert(length-2) zu Wert(length1) fortgeführt)
                        ADCFiktWert = ADCFiktWert + ADCDiff
                        GlatSum = GlatSum + ADCFiktWert
                    Next
                End If
                AMP(n) = (GlatSum / (2 * S_Glaettung.Value + 1)) * MWStg + MWOff
            End If
            AMPSum = AMPSum + AMP(n)
        Next
        AktAvgI = AMPSum / AMP.Length
        L_I.Text = AktAvgI.ToString("F2")
        ROHtoAMP = AMP
    End Function

    Private Sub Display_StandardMake(ByRef Rohwerte() As UInt16, ByVal Xmax As Double)
        'Macht eine Standardanzeige, d.h. nur eine einzige Kurve, Zeitlich von 0 bis Xmax, Ymax und Ymin werden errechnet, durchgehende Skalenlinien
        'Wertet zudem die Radiobuttons Fixed/Auto aus
        Dim AMP() As Double
        Dim AMPmax, AMPmin As Double
        Dim x As Integer

        AMP = ROHtoAMP(AktRohWerte)
        If R_ScAuto.Checked = True Then         'Auto-Skalierung - d.h. Min/Max-Werte aus AMP-Werten ermitteln
            For x = 0 To AMP.Length - 1
                If AMP(x) > AMPmax Or x = 0 Then AMPmax = AMP(x)
                If AMP(x) < AMPmin Or x = 0 Then AMPmin = AMP(x)
            Next
        Else                                    'Fixed-Skalierung - d.h. Min/Max-Werte aus eingestelltem Bereich ermitteln
            AMPmin = MWOff
            AMPmax = 1023 * MWStg + MWOff
        End If
        Display_Generate(0, Xmax, AMPmin, AMPmax, True)
        Display_AddCurve(AMP)
    End Sub

    Private Sub Display_Generate(ByVal Xmin As Double, ByVal Xmax As Double, ByVal Ymin As Double, ByVal Ymax As Double, ByVal full As Boolean)
        'mix/max Werte sind klar, full definiert, ob Skalenlinien duchgehend oder nur als Ticks an den Koordinatenlinien gezeichnet werden sollen

        Dim XRange, YRange As Double
        Dim n As Integer
        Dim bi As New Bitmap(P_I.Width, P_I.Height)
        Dim penscale As Pen
        Dim backcolor As Color
        Dim PixProXVal As Double                     'x-Skalierung Pixel/Value
        Dim SkYDiff, SkXDiff, SkDiffMin, SkDiffMax As Double   'Skalenabstand in Werten
        Dim SkVal As Double                'Laufvariable für anzuzeigende Skalenwerte
        Dim SkPix As Integer                'Laufvariable für Pixel-Koordinate der anzuzeigenden Skalenwerte
        Dim Xnull As Integer                'Die x-Koordinate der Anzeige, bei der die Null-Linie liegt (liegt evtl außerehalb der Anzeige, wenn z.B. nur ein positiver oder nur ein negativer Bereich angezeigt werden soll)

        backcolor = Color.Green
        penscale = New Pen(Color.DarkGray, 1)

        'Alte Labels disposen. Interessanterweise ist length=x+2 wenn labels von 0 bis x existieren
        For n = 0 To YLbl.Length - 2
            YLbl(n).Dispose()
        Next
        For n = 0 To XLbl.Length - 2
            XLbl(n).Dispose()
        Next


        If Ymax = Ymin Then
            YRange = 1
        Else
            YRange = Ymax - Ymin
        End If
        PixProYVal = (P_I.Height - RANDO - RANDU) / YRange                     'Y-Skalierung Pixel/Value
        Ynull = PixProYVal * Ymax + RANDO                'Ymax liegt an oberem Anzeigerand, 0-Linie entsprechend ober- oder unterhalb

        'Y-Skalendifferenzwert ermittlen, so dass Skalenabstandpixel im Min/max Rahmen liegen
        SkDiffMin = SKYABSTMIN / PixProYVal     'Minimal möglicher Skalendifferenzwert
        SkDiffMax = SkDiffMin * 2               'Maximal möglicher Skalendifferenzwert
        'Das folgende ermittelt den Skalendifferenzwert als eine 10er-Potenz von 1, 2 oder 5
        SkYDiff = 10 ^ Math.Ceiling(Math.Log10(SkDiffMin))
        'SkDiff ist jetzt nächst größere 10er-Potenz des Min-Werts (also 0.09->0.1, 0.5->1, 1->1, 6->10, 10->10, 11->100, 80->100, etc...
        'Wenn der schon passt, wird dieser mit 1. Präferenz gleich genommen (D.h. Skala wird beispielsweise 0.0, 1.0, 2.0, 3.0, etc.
        'Ist der zu groß, wird als 2. Präferenz der eins niedrigere 5er-Wert genommen
        If SkYDiff > SkDiffMax Then
            SkYDiff = SkYDiff / 2             '2. Präferenz: 5er-Schritte (z.B. 0.0, 0.5, 1.0, 1.5, 2.0, etc...)
        End If
        'Ist auch der zu groß wird als 3. Präferenz der eins niedrigere 2er-Wert genommen
        If SkYDiff > SkDiffMax Then
            SkYDiff = SkYDiff / 2.5           '3. Präferenz: 2er-Schritte (z.B. 0.0, 0.2, 0.4, 0.6, 0.8, etc)
        End If
        'Der 2er-Wert passt garantiert, da SkDiffMax = 2*SkDiffMin ist

        If Xmax = Xmin Then
            XRange = 1
        Else
            XRange = Xmax - Xmin
        End If
        PixProXVal = (P_I.Width - RANDR - RANDL) / XRange                     'X-Skalierung Pixel/Value
        Xnull = PixProXVal * Xmin + RANDL                'Xmin liegt an linkem Anzeigerand, 0-Linie entsprechend links- oder rechts davon

        'X-Skalendifferenzwert ermittlen, so dass Skalenabstandpixel im Min/max Rahmen liegen
        SkDiffMin = SKXABSTMIN / PixProXVal   'Minimal möglicher Skalendifferenzwert
        SkDiffMax = SkDiffMin * 2           'Maximal möglicher Skalendifferenzwert
        'Das folgende ermittelt den Skalendifferenzwert als eine 10er-Potenz von 1, 2 oder 5
        SkXDiff = 10 ^ Math.Ceiling(Math.Log10(SkDiffMin))
        'SkDiff ist jetzt nächst größere 10er-Potenz des Min-Werts (also 0.09->0.1, 0.5->1, 1->1, 6->10, 10->10, 11->100, 80->100, etc...
        'Wenn der schon passt, wird dieser mit 1. Präferenz gleich genommen (D.h. Skala wird beispielsweise 0.0, 1.0, 2.0, 3.0, etc.
        'Ist er zu groß wird als 2. Präferenz der eins niedrigere 5er-Wert genommen
        If SkXDiff > SkDiffMax Then
            SkXDiff = SkXDiff / 2             '2. Präferenz: 5er-Schritte (z.B. 0.0, 0.5, 1.0, 1.5, 2.0, etc...)
        End If
        'Ist auch der zu groß wird als 3. Präferenz der niedrigere 2er-Wert genommen
        If SkXDiff > SkDiffMax Then
            SkXDiff = SkXDiff / 2.5           '3. Präferenz: 2er-Schritte (z.B. 0.0, 0.2, 0.4, 0.6, 0.8, etc)
        End If
        'Der 2er-Wert passt garantiert, da SkDiffMax = 2*SkDiffMin ist

        'Image füllen
        Using gr As Graphics = Graphics.FromImage(bi)
            gr.Clear(backcolor)
            gr.SmoothingMode = SmoothingMode.AntiAlias
            'Y-Skala einzeichnen
            'Obersten Skalenwert ermitteln, Wert an der Anzeige-Obergrenze ist YMax
            SkVal = Math.Floor(Ymax / SkYDiff) * SkYDiff
            SkPix = CInt(Ynull - SkVal * PixProYVal)
            n = 0
            If full = False Then gr.DrawLine(penscale, RANDL, 0, RANDL, P_I.Height) 'Linke Skalenlinie zeichnen, wenn Skalenlinien nicht durchgaengig sein sollen
            While SkPix <= P_I.Height - RANDU                                       'Y-Saklenlinien von oben nach unten einzeichnen und entsprechende Labels anheften
                If full = False Then
                    gr.DrawLine(penscale, 0, SkPix, RANDL, SkPix)                   'Linie zeichnen - Wenn nicht durchgaengig, dann nur bis zur linken Skalenlinie
                Else
                    gr.DrawLine(penscale, 0, SkPix, P_I.Width, SkPix)               'Linie zeichnen - Wenn durchgaengig, dann von ganz links bis ganz rechts
                End If
                ReDim Preserve YLbl(n + 1)
                YLbl(n) = New Label
                YLbl(n).Text = ScaleLabel(SkVal, SkYDiff)
                YLbl(n).Width = 86
                YLbl(n).Height = 20
                YLbl(n).Left = P_I.Left - YLbl(n).Width
                YLbl(n).Top = P_I.Top + SkPix - YLbl(n).Height / 2
                YLbl(n).TextAlign = ContentAlignment.MiddleRight
                YLbl(n).Visible = True
                Me.Controls.Add(YLbl(n))
                n = n + 1
                SkVal = SkVal - SkYDiff
                SkPix = CInt(Ynull - SkVal * PixProYVal)
            End While
            'X-Skala einzeichnen
            'Linken Skalenwert ermitteln, Wert an der linken Anzeige-grenze ist XMin
            SkVal = Math.Floor(Xmin / SkYDiff) * SkYDiff
            SkPix = CInt(SkVal * PixProXVal) + RANDL
            n = 0
            If full = False Then gr.DrawLine(penscale, 0, P_I.Height - RANDU, P_I.Width, P_I.Height - RANDU) 'Untere Skalenlinie zeichnen, wenn Skalenlinien nicht durchgaengig sein sollen
            While SkPix <= P_I.Width - RANDR                                        'X-Saklenlinien von links nach rechts einzeichnen und entsprechende Labels anheften
                If full = False Then
                    gr.DrawLine(penscale, SkPix, P_I.Height, SkPix, P_I.Height - RANDU) 'Linie zeichnen - Wenn nicht durchgaengig, dann nur bis zur unteren Skalenlinie
                Else
                    gr.DrawLine(penscale, SkPix, P_I.Height, SkPix, 0)          'Linie zeichnen - Wenn durchgaengig, dann von ganz oben bis ganz unten
                End If

                ReDim Preserve XLbl(n + 1)
                XLbl(n) = New Label
                XLbl(n).Text = ScaleLabel(SkVal, SkXDiff)
                XLbl(n).Width = SkXDiff * PixProXVal
                XLbl(n).Height = 20
                XLbl(n).Left = P_I.Left + SkPix - 2
                XLbl(n).Top = P_I.Bottom
                XLbl(n).TextAlign = ContentAlignment.MiddleLeft
                XLbl(n).Visible = True
                Me.Controls.Add(XLbl(n))
                n = n + 1
                SkVal = SkVal + SkXDiff
                SkPix = CInt(SkVal * PixProXVal) + RANDL
            End While
        End Using
        P_I.Image = bi

    End Sub

    Private Sub Display_AddCurve(ByRef YVal() As Double)
        'Fügt die Werte in YVal in den Graphen ein, Ynull ist die y-Koordinate der 0-Linie
        Dim n As Integer
        Dim bi As New Bitmap(P_I.Width, P_I.Height)
        Dim penline As Pen
        Dim oldx, oldy, newx, newy As Integer

        penline = New Pen(Color.GreenYellow, 1)

        bi = P_I.Image
        'Kurve in aktuelles Bitmap einzeichnen
        Using gr As Graphics = Graphics.FromImage(bi)
            gr.SmoothingMode = SmoothingMode.AntiAlias
            'Graph einzeichnen
            oldx = RANDL
            oldy = CInt(Ynull - YVal(0) * PixProYVal)
            For n = 1 To YVal.Length - 1
                newx = CInt(Math.Round(n / YVal.Length * (P_I.Width - RANDL - RANDR), 0)) + RANDL
                newy = CInt(Ynull - YVal(n) * PixProYVal)
                gr.DrawLine(penline, oldx, oldy, newx, newy)
                oldx = newx
                oldy = newy
            Next
        End Using
        P_I.Image = bi

    End Sub

    Private Function ScaleLabel(ByVal val As Double, ByVal res As Double) As String
        'liefert Anzeige-Werte aus dem Wert val zurueck.
        '0.001 wird zu 1m, 1000 wird zu 1k, 1000000 wird zu 1M, etc.
        'Dabei wird aber die Mindestauflösung res beachtet, ansonsten wird 1000 zu 1K aber auch 1100
        'Das ist negativ, wenn man bsp. eine Skala 0,100,200,300,400,500,600,700,800,900,1000,1100,1200,1300 hat
        'Die würde ohne Midestauflösung so ausshehen: 0,100,200,300,400,500,600,700,800,900,1K,1K,1K,1K
        'Mit Midestauflösung wird sie dagegen so: 0,100,200,300,400,500,600,700,800,900,1.0K,1.1K,1.2K,1.3K
        'Ist res=0 wird immer ganz ohne Nachkommastelle ausgegeben
        Dim ch, format As String
        Dim newval, absval As Double
        Dim anzdec As Integer             'Anzahl nötiger Dezimalstellen
        anzdec = 0
        absval = Math.Abs(val)
        If absval >= 1000000000 Then
            newval = val / 1000000000
            ch = "G"
            If res > 0 Then anzdec = Ceiling(Log10(1000000000 / res))
        ElseIf absval >= 1000000 Then
            newval = val / 1000000
            ch = "M"
            If res > 0 Then anzdec = Ceiling(Log10(1000000 / res))
        ElseIf absval >= 1000 Then
            newval = val / 1000
            ch = "K"
            If res > 0 Then anzdec = Ceiling(Log10(1000 / res))
        ElseIf absval >= 0.1 Then
            newval = val
            ch = ""
            If res > 0 Then anzdec = Ceiling(Log10(1 / res))
        ElseIf absval >= 0.0001 Then
            newval = val * 1000
            ch = "m"
            If res > 0 Then anzdec = Ceiling(Log10(0.001 / res))
        ElseIf absval >= 0.0000001 Then
            newval = val * 1000000
            ch = "u"
            If res > 0 Then anzdec = Ceiling(Log10(0.000001 / res))
        ElseIf absval >= 0.0000000001 Then
            newval = val * 1000000000
            ch = "n"
            If res > 0 Then anzdec = Ceiling(Log10(0.000000001 / res))
        Else
            newval = val
            ch = ""
            anzdec = 0
        End If
        If anzdec < 0 Then anzdec = 0
        format = "F" + anzdec.ToString
        ScaleLabel = newval.ToString(format) + ch
    End Function

    Private Sub B_Null_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_Null.Click
        'Setzt aktuelles MWOff derart, dass aktuell gemesserener Strom als 0 angezeigt wird
        MWOff = MWOff - AktAvgI

    End Sub

    Private Sub R_ScAuto_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles R_ScAuto.CheckedChanged
        'Wird ausgelöst, wenn Auto an- oder auch abgewählt wird

        If imageshow = True Then                'Wenn gerade ein Bild angezeigt wird (z.B. über Load oder nach Disconnect: Neu zeichnen
            Display_StandardMake(AktRohWerte, AktRohWerte.Length * SampleRate / 1000)
        End If                                   'Wenn kein Bild angezeigt wird, sondern ein Stream laeuft - nicht neu zeichnen, Aenderung wird bei nachstem Frame automatisch aktiv
    End Sub

    Private Sub S_Glaettung_ValueChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles S_Glaettung.ValueChanged
        'Wird ausgelöst, wenn Glaettungstiefe geaendert wird

        If imageshow = True Then                'Wenn gerade ein Bild angezeigt wird (z.B. über Load oder nach Disconnect: Neu zeichnen
            Display_StandardMake(AktRohWerte, AktRohWerte.Length * SampleRate / 1000)
        End If                                   'Wenn kein Bild angezeigt wird, sondern ein Stream laeuft - nicht neu zeichnen, Aenderung wird bei nachstem Frame automatisch aktiv
    End Sub

End Class

